//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#ifndef nCore_h_
#define nCore_h_

#define NOMINMAX 1

#include <windows.h> // For now...
#include <string>

#include "nSupport.h"



#if defined(_MSC_VER) && (_MSC_VER <= 1500)
typedef signed char int8_t;
typedef unsigned char uint8_t;
typedef signed short int16_t;
typedef unsigned short uint16_t;
typedef signed int int32_t;
typedef unsigned int uint32_t;
typedef signed long long int int64_t;
typedef unsigned long long int uint64_t;
#else
#include <stdint.h>
#endif

static const char version_string[] = "lossyWAV ";
//static const char CRLF[] = "\n";
#define CRLF std::endl

static const int EXPONENT_BIAS = 1024;

static const int PRECALC_ANALYSES = 6;
static const int BITS_TO_CALCULATE = 32;

static const int WORD_SIZE = 2;
static const int WORD_SHIFT = 1;

static const int SINGLE_SIZE = 4;
static const int SINGLE_SHIFT = 2;

static const int DOUBLE_SIZE = 8;
static const int DOUBLE_SHIFT = 3;

static const int COMPLEX_SIZE = 16;
static const int COMPLEX_SHIFT = 4;

static const int INTEGER_SIZE = 4;
static const int INTEGER_SHIFT = 2;

static const int CARDINAL_SIZE = 4;
static const int CARDINAL_SHIFT = 2;

static const int INT64_SIZE = 8;
static const int INT64_SHIFT = 3;

static const int POINTER_SIZE = 4;
static const int POINTER_SHIFT = 2;

static const double e = 2.7182818284590452353602874713527;                    // e
static const double OneOverPi = 0.31830988618379067153776752674503;           // 1/Pi
static const double OneOverTwoPi = 0.15915494309189533576888376337251;		  // 1/(2*Pi)
static const double HalfPi = 1.5707963267948966192313216916398;               // Pi/2
static const double Pi = 3.1415926535897932384626433832795;                   // Pi
static const double TwoPi = 6.2831853071795864769252867665590;				  // Pi*2

static const double OneOverRoot2 = 0.70710678118654752440084436210485;        // 2^-0.5
static const double RootTwo = 1.4142135623730950488016887242097;              // 2^0.5
static const double log2_10 = 3.3219280948873623478703194294894;              // 1/log10(2)
static const double log10_2 = 0.30102999566398119521373889472449;             // log10(2)
static const double log10_2x20 = 6.205999132796239042747778944899;            // 20 * log10(2)
static const double MaxDb = 5623413251903.4908039495103977648;				  // 10^(255/20)
static const double GoldenRatio = 1.61803398874989484820458683436564;		  // Golden Ratio
static const double Second_div_Day = 1.0/(24*60*60);	                      // 1/(24*60*60)

//============================================================================
// Core settings relating to WAV limits and processing limit.
//============================================================================
static const unsigned int MAX_WAVE_SIZE = 4294967278u;
static const int MAX_CHANNELS = 8;
static const int MAX_BLOCK_BITS = 12;
static const unsigned int MAX_BLOCK_SIZE = 1 << MAX_BLOCK_BITS;
static const unsigned int BUFFER_SIZE = MAX_CHANNELS * MAX_BLOCK_SIZE * INTEGER_SIZE;
static const unsigned int CHANNEL_BYTE_SIZE = MAX_BLOCK_SIZE * INTEGER_SIZE;

//============================================================================
// MAX_FFT_LENGTH **MUST** be double (or greater) MAX_BLOCK_SIZE (!!)
//============================================================================
static const int MAX_FFT_BIT_LENGTH = MAX_BLOCK_BITS + 1;
static const unsigned int MAX_FFT_LENGTH = 1 << MAX_FFT_BIT_LENGTH;
static const unsigned int MAX_FFT_LENGTH_HALF = MAX_FFT_LENGTH / 2;
static const unsigned int MAX_FFT_LENGTH_RECIP = 1 / MAX_FFT_LENGTH;

static const int LONG_FFT_LENGTH = 1024;      // for 44.1/48kHz content
static const int ADDED_FFT_LENGTH_512 = 512;  // for 44.1/48kHz content
static const int ADDED_FFT_LENGTH_256 = 256;  // for 44.1/48kHz content
static const int ADDED_FFT_LENGTH_128 = 128;  // for 44.1/48kHz content
static const int SHORT_FFT_LENGTH = 64;	   // for 44.1/48kHz content
static const int IMPULSE_FFT_LENGTH = 32;	   // for 44.1/48kHz content

static const int LONG_FFT_BIT_LENGTH = 10;
static const int SHORT_FFT_BIT_LENGTH = 6;

static const int MAX_MUL = 8;                 // shift for 352.4/384kHz content
static const int MAX_SHIFT = 3;			   // shift for 352.4/384kHz content

static const unsigned int MAX_HIST_FFT_SHORT_LENGTH = SHORT_FFT_LENGTH << MAX_SHIFT;

//============================================================================
// Core Settings.
//============================================================================
static const int STATIC_MINIMUM_BITS_TO_KEEP = 6;

//============================================================================
// Frequency range over which to check for lowest amplitude signal
//============================================================================
static const int LOWER_FREQ_LIMIT = 20;
static const int MID_FREQ_LIMIT = 3700;
static const int UPPER_FREQ_LIMIT = 16000;
static const int HUMAN_FREQ_LIMIT = 20000;


//============================================================================
// Types
//============================================================================

struct tDComplex
{
	//tDComplex(double re = 0.0, double im = 0.0) : Re(re), Im(im) {}
	void set(double re, double im) { Re = re; Im = im; }
	double Re, Im;
};

struct FFT_results_rec
{
	short analysis;
	short btr;
	int start;
	double spreading;
};

struct channel_data_rec
{
	int maximum_bits_to_remove;
	int bits_to_remove;
	int bits_lost;
	int clips;
	FFT_results_rec min_FFT_result;
};

struct FFT_Data_Rec
{
	int bit_length;
	int bit_shift_from_max;
	int length;
	int length_m1;
	int length_half;
	int length_half_m1;
	float length_recip;
	float length_half_recip;
	float threshold_shift;
};

typedef int SingleChannelCodecBlock[ MAX_BLOCK_SIZE ];
typedef SingleChannelCodecBlock MultiChannelCodecBlock[ MAX_CHANNELS ];
typedef SingleChannelCodecBlock* MultiChannelCodecBlockPtr;
typedef double Half_FFT_Double_Array[ MAX_FFT_LENGTH_HALF + 2 ];
typedef Half_FFT_Double_Array Half_FFT_Channel_Double_Array[ MAX_CHANNELS ];

static const int HFDA_Size = sizeof(Half_FFT_Double_Array);
static const int HFCDA_Size = sizeof(Half_FFT_Channel_Double_Array);

//============================================================================
// Array for in-place FFT analysis.
//============================================================================
extern struct FFT_Array_type
{
	union
	{
		tDComplex DComplex[ MAX_FFT_LENGTH ];
		double DReal[ MAX_FFT_LENGTH * 2 ];
	};
}
FFT_Array;

//============================================================================
// Processed FFT Result storage (FFT procedure is in-place).
//============================================================================
extern double FFT_raw_result[ MAX_FFT_LENGTH ];
extern double FFT_last_result[ MAX_FFT_LENGTH ];
extern double FFT_skew_result[ MAX_FFT_LENGTH ];

//============================================================================
// All audio data
//============================================================================
extern struct BlockData_type
{
	MultiChannelCodecBlock Data[ 4 ];
	MultiChannelCodecBlock BTRD;
	MultiChannelCodecBlock CORR;
	double Channel_Log2_RMS[ MAX_CHANNELS ];
	struct
	{
		int Prev, Last, This, Next;
	}
	Size;
	int CHANNEL_OFFSET[ MAX_CHANNELS ];
	MultiChannelCodecBlockPtr Pointer[ 4 ];
}
BlockData;

//============================================================================
// Lookup table for integer reciprocal.
//============================================================================
extern double OneOver[ MAX_FFT_LENGTH + 2 ];

//============================================================================
// Global variables used by the Units.
//============================================================================
static const unsigned char REVERSED_NIBBLES[ 16 ] = { 0x00, 0x08, 0x04, 0x0C, 0x02, 0x0A, 0x06, 0x0e, 0x01, 0x09, 0x05, 0x0D, 0x03, 0x0B, 0x07, 0x0F };
static const char HEXCHARLOOKUP[ 16 ] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

static const float SINGLE_HALF = 0.50;
static const float SINGLE_THREE_QUARTERS = 0.75;
static const float SINGLE_ONE = 1.0;
static const float SINGLE_ONE_AND_A_HALF = 1.50;
static const double db_MaxDB = MaxDb;
static const double db_lg2x20 = log10_2x20;

//typedef HANDLE tHandle;


extern struct timer_type
{
	LARGE_INTEGER Frequency;
	double Period;
	LARGE_INTEGER start;
	LARGE_INTEGER Stop;
	double Elapsed;
	int64_t DelayEnd;
	int64_t Packer;
	time_t StartTime;
	double StopTime;
	double DelayTime;
}
timer;

extern struct Global_type
{
	int analysis_number;
	struct
	{
		int bits[ PRECALC_ANALYSES + 1 ];
		int length[ PRECALC_ANALYSES + 1 ];
	}
	analysis;

	FFT_Data_Rec FFT;
	int channel;
	int Codec_Block_size;
	double Codec_Block_size_recip;
	double Codec_Block_duration;
	int channels;
	int Codec_Block_bits;
	int Codec_Block_bit_shift;
	int sample_rate;
	int bits_per_sample;
	int bytes_per_sample;
	unsigned int wave_size;
	unsigned int total_samples;
	int lower_freq_limit;
	int upper_freq_limit;
	int blocks_processed;
	int bits_removed;
	int bits_lost;
	int last_print;
	int output_blocks;
	double sample_rate_recip;
	double blocks_processed_recip;
	double processing_rate;
	int64_t samples_processed;
	int64_t clipped_samples;
	double samples_processed_recip;
	int jump_location;
}
Global;

extern struct parameters_type
{
	std::string logfilename;
	std::string wavName;
	std::string stdinname;
	std::string lossyName;
	std::string lwcdfName;
	std::string WavInpDir;
	std::string WavOutDir;
	bool detail;
	bool verbosity;
	bool warnings;
	bool forcing;
	bool correction;
	bool checking;
	bool merging;
	bool silent;
	bool linkchannels;
	bool writetolog;
	bool STDINPUT;
	bool STDOUTPUT;
	bool perchannel;
	bool bitdist;
	int spread;
	bool blockdist;
	bool sampledist;
	bool logfileopened;
	bool midside;
	bool dccorrect;
	bool parameters_checked;
	int adaptive_shaping;
	bool adaptive_shaping_warp;
	int Static;
	double dynamic;
	double quality;
	double scaling;
	double shaping;
	int analyses;
	int priority;
	int help;
	int limit;
	int maxclips;
	int underlap;
	bool freqdist;
	bool postanalyse;
	bool histogram;
	int width;
	bool longdist;
}
parameters;

extern struct settings_type
{
	double noise_threshold_shift_minimum;
	double noise_threshold_shift_average;
	double dccorrect_multiplier;
	double dynamic_minimum_bits_to_keep;
	double dynamic_maximum_bits_to_remove;
	int static_minimum_bits_to_keep;
	int static_maximum_bits_to_remove;
	double scaling_factor;
	double scaling_factor_inv;
	double quality_double;
	double quality_fraction;
	int quality_integer;
	int maximum_clips_per_channel;
	int FFT_analysis_switches;
	int raw_result_short;
	int raw_result_sgns;
}
settings;

extern struct process_type
{
	struct
	{
		double old_minimum;
		double new_minimum;
		double alt_average;
		double Masked_Minimum;
	}
	FFT_spreading[ PRECALC_ANALYSES + 1 ][ MAX_CHANNELS ];

	double FFT_underlap_length[ MAX_FFT_BIT_LENGTH + 1 ];
	int analysis_blocks[ MAX_FFT_BIT_LENGTH + 1 ];
	int end_overlap_length[ MAX_FFT_BIT_LENGTH + 1 ];
	int actual_analysis_blocks_start[ MAX_FFT_BIT_LENGTH + 1 ];
	int Analyses_Completed[ PRECALC_ANALYSES + 1 ];
	int Old_Min_Used_History[ PRECALC_ANALYSES + 1 ][ MAX_FFT_LENGTH_HALF ];
	int New_Min_Used_History[ PRECALC_ANALYSES + 1 ][ MAX_FFT_LENGTH_HALF ];
	int Alt_Ave_Used[ PRECALC_ANALYSES + 1 ];
	int Old_Min_Used[ PRECALC_ANALYSES + 1 ];
	int New_Min_Used[ PRECALC_ANALYSES + 1 ];
	int Over_Static[ PRECALC_ANALYSES + 1 ];
	int Over_Dynamic[ PRECALC_ANALYSES + 1 ];
	void (* Post_Analysis[ PRECALC_ANALYSES + 1 ])() ;
	int old_min_bin;
	int new_min_bin;
	int dynamic_maximum_bits_to_remove;
	int Analyses_Performed;
}
process;

extern struct results_type
{
	FFT_results_rec saved_FFT_results[ MAX_CHANNELS ][ PRECALC_ANALYSES + 1 ];
	int minima[ MAX_CHANNELS ][ PRECALC_ANALYSES + 3 ];
	FFT_results_rec this_FFT;
}
results;

extern struct history_type
{
	double DATA_results[ MAX_CHANNELS ][ MAX_HIST_FFT_SHORT_LENGTH / 2 + 2 ];
	double BTRD_results[ MAX_CHANNELS ][ MAX_HIST_FFT_SHORT_LENGTH / 2 + 2 ];
	double CORR_results[ MAX_CHANNELS ][ MAX_HIST_FFT_SHORT_LENGTH / 2 + 2 ];
	double Last_Results[ MAX_CHANNELS ][ MAX_HIST_FFT_SHORT_LENGTH / 2 + 2 ];
	FFT_Data_Rec FFT;
	int CHANNEL_OFFSET[ MAX_CHANNELS ];
	int Histogram_DATA[ 1026 ];
	int Histogram_BTRD[ 1026 ];
	int Histogram_CORR[ 1026 ];
	double Histogram_Multiplier;
	int Histogram_Length;
	int Histogram_Offset;
}
history;

extern struct LongDist_type
{
	double DATA_results[ MAX_CHANNELS ][ MAX_FFT_LENGTH_HALF + 2 ];
	double BTRD_results[ MAX_CHANNELS ][ MAX_FFT_LENGTH_HALF + 2 ];
	double CORR_results[ MAX_CHANNELS ][ MAX_FFT_LENGTH_HALF + 2 ];
	FFT_Data_Rec FFT;
	int CHANNEL_OFFSET[ MAX_CHANNELS ];
}
LongDist;

extern struct SGNSFFT_type
{
	FFT_Data_Rec FFT;
	int CHANNEL_OFFSET[ MAX_CHANNELS ];
	double FFT_results_short[ MAX_FFT_LENGTH_HALF + 2 ];
	double FFT_results_long[ MAX_FFT_LENGTH_HALF + 2 ];
	double Last_Results[ MAX_CHANNELS ][ MAX_FFT_LENGTH_HALF + 2 ];
}
SGNSFFT;

extern struct strings_type
{
	std::string parameter;
	std::string datestamp;
	std::string Size;
	std::string time;
	std::string Elapsed;
	std::string estimate;
	std::string version;
	std::string version_short;
}
strings;

extern struct version_type
{
	union
	{
		struct { unsigned int ReleaseBuild, MajorMinor; };
		struct { unsigned short Build, Release, Minor, Major; };
	};
}
version;

extern struct PowersOf_type
{
	/*int64_t*/ long TwoInt[ 64 ];
	/*int64_t*/ long TwoM1[ 64 ];
	range_array< double, -1024, 1023 > Two;
	range_array< double, -308, 307 > Ten;
}
PowersOf;

extern struct Stats_type
{
	int bits_removed[ MAX_CHANNELS ][ 34 ];
}
Stats;

void lossyWAVerror(std::string lwe_string, int lwe_value);
void lossyWAVwarning(std::string lww_string);
void gettimer();
void setpriority(int spv);
void ndelay();
std::string CardinalToHex(unsigned int cth_c);
void time_string_make(std::string& tsms, double tsmt);
void size_string_make(std::string& ssms, double ssmt);
void Set_FFT_Data_Rec(FFT_Data_Rec& FFTDataRec, int SFDR_Bits);
void write_unitname_and_datapointers(std::string UnitName, void* UserData1Pointer, void* UserData2Pointer, unsigned int UserData2Size);

#endif
