//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//    Initial translation to C++ from Delphi
//    by Tyge Lvset (tycho), Aug. 2012
//==============================================================================

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <ctime>

#include "nWav.h"
#include "nCore.h"
#include "nParameter.h" // filemode
#include "nMaths.h"

namespace {

static const char UnitName[] = "nWav";

const int CUEChunkMaxSize = 65536;
const int FACTChunkMaxSize = 65536;
const int FMTChunkMaxSize = 18 + 22;


typedef char tChunkID[ 4 ];

struct tChunkHeader
{
	union
	{
		struct
		{
			tChunkID id;
			unsigned int Size;
		};
		struct
		{
			unsigned int CNum;
		};
	};
};


struct tChunkMapRecord
{
	tChunkHeader Header;
	int CPtr;
	int Packer;
};

struct tWAVHeader
{
	tChunkHeader Header; // "RIFF"
	tChunkID RIFFType;   // "WAVE"
	int Packer;
};

struct tFMTChunk
{
	tChunkHeader Header; // fmt
	unsigned short wFormatTag;
	unsigned short wChannels;
	unsigned int nSamplesPerSec;
	unsigned int nAvgBytesPerSec;
	unsigned short nBlockAlign;
	unsigned short wBitsPerSample;
	unsigned short cbSize;

	struct
	{
		union
		{
			unsigned short wValidBitsPerSample;
			unsigned short wSamplesPerBlock;
			unsigned short wReserved;
		};
	}
	Samples;

	unsigned int dwChannelMask;
	unsigned short SubFormat_Word[ 4 ];
	unsigned char SubFormat_Byte[ 8 ];
};

struct tFACTChunk
{
	tChunkHeader Header; // "fact"
	char Data[ 65536 ];
};


struct tDATAChunk
{
	tChunkHeader Header; // "data"
};


struct tlossyWAVChunk
{
	tWAVHeader WAV;
	tFMTChunk FMT;
	tFACTChunk FACT;
	tDATAChunk Data;
};


struct tWaveBuffer
{
	union
	{
		short w[ BUFFER_SIZE / 2 ];
		struct { unsigned char lo_b; short hi_w; }
		      t[ BUFFER_SIZE / 3 ];
		int   l[ BUFFER_SIZE / 4 ];
	};
};

static const int chunkspace_array_size = 2097152;

tWaveBuffer WAVBuff;
tWaveBuffer BtrdBuff;
tWaveBuffer Corrbuff;
tWaveBuffer MergeBuff;

unsigned char* chunkspace;
tChunkMapRecord chunkmap[ 256 ];

std::fstream WAVEFile, Btrdfile, Corrfile;

std::string DateTimeString;

tChunkHeader ChunkHeader;
tlossyWAVChunk WAVEChunk;
tlossyWAVChunk BTRDChunk;
tlossyWAVChunk CORRChunk;

unsigned int filesOpened;
unsigned int readresult;
unsigned int writeresult;

unsigned int BUFFER_SIZEread;                   // number of byte that make up for inbuff
unsigned int BUFFER_SIZEwrite;					// number of byte that make up for outbuff
unsigned int nrOfByteInOneBlockInBuff;			// number of byte that make up a complete block in inBuff and outBuff.
unsigned int nrOfBlocksInReadBuff;
unsigned int nrOfBlocksInWriteBuff;

unsigned int nrOfFullBlocksReadIntoInBuff;		// with the last BlockRead: how many full blocks managed it into inBuff ?
unsigned int nrOfByteOfTheLastIncompleteBlock;	// when EOF was encountered with the last BlockRead: the number of byte that didn't make it
												// into the last full block
unsigned int nrOfBlockInBuffNotYetFetched;		// the next number of the block in inBuff that was not fetched yet by readNextSampleBlock.
												// 0: no data in inBuff;
unsigned int nrOfByteFilledInBtrdBuff;			// how many byte have managed it into outBuff that are not written to the outputfile yet?
unsigned int nrOfByteFilledInCorrBuff;			// how many byte have managed it into outBuff that are not written to the outputfile yet?

unsigned int sampleByteLeftToRead;
unsigned int samplebytesread;
unsigned int readbyteCount;
unsigned int byteslefttoread;
unsigned int thisreadbyteCount;

unsigned int writebytecount;
unsigned int thiswritebytecount;
unsigned int firstsamplesread;
unsigned int lastwinerror;
unsigned int lastioresult;
unsigned int datachunkpos;
unsigned int ThisFilePos;

unsigned int retries;
bool CantReadInWavAnyMore;

unsigned int numchunks = 0;
unsigned int endofchunkmap;

unsigned int BUFFER_SIZEtouse;
int last_find_text = -1;

void (* readtransfer)(PByte pB, PByte pEndB);
void (* writetransfer)(MultiChannelCodecBlock& outputcodecblock, PByte pB);
int Unit_Data_Packer2[ 4 ];

} // namespace



void combinelossybuffers(int bytes_in_buffer)
{
	int cb_i, cb_j;
	int this_sample, btrd_sample, corr_sample;
	cb_i = bytes_in_buffer / ((int) Global.bytes_per_sample);

	switch (Global.bytes_per_sample)
	{
		case 1:
			lossyWAVerror("Invalid bitdepth.", 0x12);
			break;

		case 2:
			for (cb_j = 0; cb_j < cb_i; ++cb_j)
			{
				MergeBuff.w[cb_j] = nround(BtrdBuff.w[cb_j] * settings.scaling_factor_inv + Corrbuff.w[cb_j]);
			}

			break;

		case 3:
			for (cb_j = 0; cb_j < cb_i; ++cb_j)
			{
				btrd_sample = BtrdBuff.t[cb_j].hi_w * 256 | BtrdBuff.t[cb_j].lo_b;
				corr_sample = Corrbuff.t[cb_j].hi_w * 256 | BtrdBuff.t[cb_j].lo_b;
				this_sample = nround(btrd_sample * settings.scaling_factor_inv + corr_sample);
				MergeBuff.t[cb_j].lo_b = ((unsigned char) this_sample & 255);
				MergeBuff.t[cb_j].hi_w = this_sample / 63356;
			}

			break;

		case 4:
			for (cb_j = 0; cb_j < cb_i; ++cb_j)
			{
				MergeBuff.l[cb_j] = nround(BtrdBuff.l[cb_j] * settings.scaling_factor_inv + Corrbuff.l[cb_j]);
			}

			break;
	}
}


void OpenFile(std::fstream& thisfile, std::string thisname, unsigned int thisfilemode, unsigned int orvalue)
{
	switch (thisfilemode)
	{
		case 0:
            //Reset(thisfile, 1);
            thisfile.open(thisname.c_str(), std::ios::in | std::ios::binary);
			break;

		case 2:
            //Rewrite(thisfile, 1);
            thisfile.open(thisname.c_str(), std::ios::out | std::ios::binary);
			break;

		default:
			lossyWAVerror("Invalid FileMode!", 0x12);
	}

    if (!thisfile.good())
	{
		return;
	}

	filesOpened |= orvalue;
}


void wavIOExitProc(std::string wavIOString, int wavIOCode)
{
	unsigned int wi;

	for (wi = 0; wi < numchunks; ++wi)
	{
        ErrOutput << chunkmap[wi].Header.id << ',';
	}

    ErrOutput << CRLF << wavIOString << CRLF;
	closeWavIO();
	ExitCode = wavIOCode;

	throw (ExitCode);
}


unsigned int readfromWAVEfile(void* buffpointer, unsigned int bytestoread)
{
	unsigned int result;
	readbyteCount = 0;
	retries = 0;

	if (parameters.STDINPUT)
	{
		while ((readbyteCount < bytestoread) && (! CantReadInWavAnyMore))
		{
			byteslefttoread = bytestoread - readbyteCount;
            // if (ReadFile(std::cin, ((Pointer)((unsigned int) buffpointer) + readbyteCount), byteslefttoread, thisreadbyteCount, NULL))
            std::cin.read(static_cast<char *>(buffpointer) + readbyteCount, byteslefttoread);
            thisreadbyteCount = std::cin.gcount();

            if (std::cin.good())
            {
				lastwinerror = 0;
			}
			else
			{
				lastwinerror = GetLastError();
			}

			if (thisreadbyteCount == 0)
			{
				++ retries;
                timer.DelayTime = PowersOf.Two[std::min(7u, retries) -11];
				ndelay();
			}
			else
			{
				retries = 0;
			}

			if ((lastwinerror == ERROR_BROKEN_PIPE) || (retries == 32))
			{
				CantReadInWavAnyMore = true;
			}

            readbyteCount += thisreadbyteCount;
		}
	}
	else
	{
        //blockread(WAVEFile, buffpointer, bytestoread, readbyteCount);
        WAVEFile.read((char *) buffpointer, bytestoread);
        readbyteCount = WAVEFile.gcount();
        lastioresult = WAVEFile.good() ? 0 : 1;

		if ((lastioresult != 0) || (readbyteCount < bytestoread))
		{
			CantReadInWavAnyMore = true;
		}
	}

	result = readbyteCount;
	return result;
}


unsigned int WordAlign(unsigned int WAvalue)
{
	return WAvalue + (WAvalue & 1);
}


bool WriteChunks(std::ostream& outputfile)
{
	bool result = false;
	unsigned int thischunk;

	if ((parameters.merging || parameters.STDOUTPUT) && (WAVEChunk.FACT.Header.Size != 0))
	{
		WAVEChunk.WAV.Header.Size -= (WAVEChunk.FACT.Header.Size + 8);
	}

    //BlockWrite(outputfile, WAVEChunk.WAV, 12);
    outputfile.write((const char *)&WAVEChunk.WAV, 12);

	if (ioresult != 0)
	{
		return result;
	}

	for (thischunk = 0; thischunk < numchunks; ++thischunk)
	{
		if (chunkmap[thischunk].CPtr == -1)
		{
			switch (chunkmap[thischunk].Header.CNum)
			{
				case 0x20746D66:
					{
                        //BlockWrite(outputfile, WAVEChunk.FMT, WordAlign(WAVEChunk.FMT.Header.Size + 8));
                        outputfile.write((const char*) &WAVEChunk.FMT, WordAlign(WAVEChunk.FMT.Header.Size + 8));

						if (ioresult != 0)
						{
							wavIOExitProc("Error writing ''fmt '' chunk.", 0x12);
						}

						if ((! parameters.merging) && (WAVEChunk.FACT.Header.Size != 0) && (! parameters.STDOUTPUT))
						{
                            //BlockWrite(outputfile, WAVEChunk.FACT, WordAlign(WAVEChunk.FACT.Header.Size + 8));
                            outputfile.write((const char*) &WAVEChunk.FACT, WordAlign(WAVEChunk.FACT.Header.Size + 8));

							if (ioresult != 0)
							{
								wavIOExitProc("Error writing ''fact'' chunk.", 0x12);
							}
						}
					}
					break;

				case 0x74636166:
					{
					}
					break;
			}
		}
		else
		{
            //BlockWrite(outputfile, chunkmap[thischunk].Header, 8);
            outputfile.write((const char*) &chunkmap[thischunk].Header, 8);

			if (ioresult != 0)
			{
				wavIOExitProc(std::string("Error writing ''") + chunkmap[thischunk].Header.id + "'' chunk.", 0x12);
			}

            //BlockWrite(outputfile, chunkspace[chunkmap[thischunk].CPtr], WordAlign(chunkmap[thischunk].Header.Size));
            outputfile.write((const char*) &chunkspace[chunkmap[thischunk].CPtr], WordAlign(chunkmap[thischunk].Header.Size));

			if (ioresult != 0)
			{
				wavIOExitProc(std::string("Error writing ''") + chunkmap[thischunk].Header.id + "'' chunk.", 0x12);
			}
		}
	}

	if (! parameters.STDOUTPUT)
	{
        datachunkpos = outputfile.tellp();
	}

    //BlockWrite(outputfile, WAVEChunk.Data, 8);
    outputfile.write((const char*) &WAVEChunk.Data, 8);

	if (ioresult != 0)
	{
		return result;
	}

	result = true;
	return result;
}


void RewriteDataChunk(std::ostream& outputfile)
{
    ThisFilePos = outputfile.tellp();
    outputfile.seekp(0);
	WAVEChunk.WAV.Header.Size = ThisFilePos - 8;
    //BlockWrite(outputfile, WAVEChunk.WAV, 8, writeresult);
    outputfile.write((const char*) &WAVEChunk.WAV, 8);
    outputfile.seekp(datachunkpos);
	WAVEChunk.Data.Header.Size = samplebytesread;
    //BlockWrite(outputfile, WAVEChunk.Data, 8, writeresult);
    outputfile.write((const char*) &WAVEChunk.Data, 8);
}


void ReadChunks(std::istream& thisfile, tlossyWAVChunk& thischunk)
{
	std::memset((void*) &thischunk, 0, sizeof(tlossyWAVChunk));
	numchunks = 0;
	endofchunkmap = 0;

    //blockread(thisfile, ChunkHeader, 8, readresult);
    thisfile.read((char*) &ChunkHeader, 8);
    readresult = thisfile.gcount();

	if (readresult != 8)
	{
		wavIOExitProc("Error Reading WAV file Information.", 0x11);
	}

	if (ChunkHeader.id == "RIFF")
	{
		std::memcpy(&thischunk.WAV, &ChunkHeader, 8);
        //blockread(thisfile, thischunk.WAV.RIFFType, 4, readresult);
        thisfile.read(thischunk.WAV.RIFFType, 4);
        readresult = thisfile.gcount();

		if (readresult != 4)
		{
			wavIOExitProc("Error Reading WAV file Information.", 0x11);
		}
	}
	else
	{
		wavIOExitProc("File type is not RIFF.", 0x12);
	}

	if (thischunk.WAV.RIFFType != "WAVE")
	{
		wavIOExitProc("RIFF type is not WAVE.", 0x12);
	}

	do
	{
        //blockread(thisfile, ChunkHeader, 8, readresult);
        thisfile.read((char*) &ChunkHeader, 8);
        readresult = thisfile.gcount();

		if (readresult != 8)
		{
			wavIOExitProc("Error Reading WAV file Information.", 0x11);
		}

		switch (ChunkHeader.CNum)
		{
			case 0x20746D66: // FMT
				{
					chunkmap[numchunks].Header = ChunkHeader;
					chunkmap[numchunks].CPtr = -1;
					++ numchunks;
					std::memcpy(&thischunk.FMT, &ChunkHeader, 8);

					switch (thischunk.FMT.Header.Size)
					{
						case 16: case 18: case 40:
							{
							}
							break;

						default:
							wavIOExitProc("FMT Chunk incorrect size.", 0x12);
					}

                    //blockread(thisfile, thischunk.FMT.wFormatTag, WordAlign(thischunk.FMT.Header.Size), readresult);
                    thisfile.read((char*) &thischunk.FMT.wFormatTag, WordAlign(thischunk.FMT.Header.Size));
                    readresult = thisfile.gcount();

					if (readresult != WordAlign(thischunk.FMT.Header.Size))
					{
						wavIOExitProc("Error reading FMT chunk.", 0x12);

//						if (thischunk.FMT.wFormatTag == 1 && thischunk.FMT.wBitsPerSample == 40)
//						{
//							thischunk.FMT.Samples.wValidBitsPerSample = thischunk.FMT.wBitsPerSample;
//							thischunk.FMT.wFormatTag = 0xFFFE;
//							thischunk.FMT.SubFormat_Word[0] = 1;
//						}
					}
				}
				break;

			case 0x61746164: // Data
				{
					std::memcpy(&thischunk.Data, &ChunkHeader, 8);
				}
				break;

			default:
				chunkmap[numchunks].Header = ChunkHeader;
				chunkmap[numchunks].CPtr = endofchunkmap;

				if (endofchunkmap > chunkspace_array_size)
				{
					wavIOExitProc("Too much WAV information before ''data'' chunk!", 0x12);
				}

                //blockread(thisfile, chunkspace[endofchunkmap], WordAlign(ChunkHeader.Size), readresult);
                thisfile.read((char*) &chunkspace[endofchunkmap], WordAlign(ChunkHeader.Size));
                readresult = thisfile.gcount();

				if (readresult != WordAlign(ChunkHeader.Size))
				{
					wavIOExitProc(std::string("Error Reading ''") + ChunkHeader.id + "'' Chunk.", 0x12);
				}

				endofchunkmap = endofchunkmap + WordAlign(ChunkHeader.Size);

				if (chunkmap[numchunks].Header.id == "fact")
				{
					thischunk.FACT.Header = chunkmap[numchunks].Header;

					if (thischunk.FACT.Header.Size < FACTChunkMaxSize)
					{
						std::memcpy(&thischunk.FACT.Data, &chunkspace[chunkmap[numchunks].CPtr], thischunk.FACT.Header.Size);
					}

					if (std::string(&thischunk.FACT.Data[0], 8) != "lossyWAV")
					{
						thischunk.FACT.Header.Size = 0;
					}
					else
					{
						chunkmap[numchunks].CPtr = -1;
					}
				}

				++ numchunks;
		}
	}
	while (!(ChunkHeader.id == "data"));

	if ((thischunk.FMT.wFormatTag != 1) && ((thischunk.FMT.wFormatTag == 0xFFFE) && (thischunk.FMT.SubFormat_Word[0] != 1)))
	{
		wavIOExitProc("WAV Data is not uncompressed PCM.", 0x12);
	}
}

//int find_text(std::string find_str, std::string source_str); // Not needed


void MergeFiles()
{
	unsigned int thisread;
	unsigned int numreads;
	unsigned int lastread;
	unsigned int readcount;
	int this_error = 0;
	std::string temp_str;
	filesOpened = 0;
	OpenFile(Btrdfile, parameters.WavInpDir + parameters.lossyName, 0, 2);
	ReadChunks(Btrdfile, BTRDChunk);
	OpenFile(Corrfile, parameters.WavInpDir + parameters.lwcdfName, 0, 2);
	ReadChunks(Corrfile, CORRChunk);
    ErrOutput << CRLF << "Merging: " << parameters.lossyName << CRLF;
    ErrOutput << "       & " << parameters.lwcdfName << CRLF;

	if (BTRDChunk.WAV.Header.Size != CORRChunk.WAV.Header.Size)
	{
		lossyWAVerror("Files not same length.", 0x12);
	}

	if (BTRDChunk.Data.Header.Size != CORRChunk.Data.Header.Size)
	{
		lossyWAVerror("Data not same length.", 0x12);
	}

	if (BTRDChunk.FACT.Header.Size == 0)
	{
		lossyWAVerror("FACT Data missing (lossy.wav).", 0x12);
	}

	if (CORRChunk.FACT.Header.Size == 0)
	{
		lossyWAVerror("FACT Data missing (lwcdf.wav).", 0x12);
	}

	if (BTRDChunk.FACT.Data != CORRChunk.FACT.Data)
	{
		lossyWAVerror("FACT Data mismatch.", 0x12);
	}

	if (std::string(BTRDChunk.FACT.Data[0], 8) != "lossyWAV")
	{
		lossyWAVerror("Not a lossyWAV FACT chunk.", 0x12);
	}

	if ((BTRDChunk.FMT.Header.Size != CORRChunk.FMT.Header.Size) || (BTRDChunk.FMT.wFormatTag != CORRChunk.FMT.wFormatTag) ||
		(BTRDChunk.FMT.wChannels != CORRChunk.FMT.wChannels) || (BTRDChunk.FMT.nSamplesPerSec != CORRChunk.FMT.nSamplesPerSec) ||
		(BTRDChunk.FMT.nAvgBytesPerSec != CORRChunk.FMT.nAvgBytesPerSec) || (BTRDChunk.FMT.nBlockAlign != CORRChunk.FMT.nBlockAlign) ||
		(BTRDChunk.FMT.wBitsPerSample != CORRChunk.FMT.wBitsPerSample))
	{
		lossyWAVerror("FMT Data mismatch.", 0x12);
	}

	Global.bytes_per_sample = (BTRDChunk.FMT.wBitsPerSample + 7) >> 3;

	if (Global.bytes_per_sample < 2)
	{
		lossyWAVerror(std::string("Invalid bitdepth: ") + IntToStr(BTRDChunk.FMT.wBitsPerSample), 0x12);
	}

	lastread = BTRDChunk.Data.Header.Size % BUFFER_SIZE;
	numreads = BTRDChunk.Data.Header.Size / BUFFER_SIZE + ((unsigned int) lastread > 0);
	std::memcpy(&WAVEChunk, &BTRDChunk, sizeof(BTRDChunk));
	OpenFile(WAVEFile, parameters.WavOutDir + parameters.wavName, 2, 1);

	if (! WriteChunks(WAVEFile))
	{
		lossyWAVerror("Writing to output file.", 0x12);
	}

	std::string source(&WAVEChunk.FACT.Data[0]), key("--scale ");
	size_t pos = source.find(key);
	
	if (pos != std::string::npos)
	{
		settings.scaling_factor = std::atof(&source[pos + key.length()]);
	}
	else
	{
		settings.scaling_factor = 1.0;
	}

	if (settings.scaling_factor <= 0.0)
	{
		lossyWAVerror(std::string("Invalid scale value: ") + source.substr(pos + 8, 8), 0x12);
	}

	settings.scaling_factor_inv = 1.0 / settings.scaling_factor;
	BUFFER_SIZEtouse = BUFFER_SIZE - (BUFFER_SIZE % (Global.bytes_per_sample * BTRDChunk.FMT.wChannels));

	for (readcount = 0; readcount < numreads; ++readcount)
	{
		thisread = std::min(BUFFER_SIZEtouse, BTRDChunk.Data.Header.Size - readcount * BUFFER_SIZEtouse);
        //blockread(Btrdfile, BtrdBuff, thisread, readresult);
        Btrdfile.read((char*) &BtrdBuff, thisread);
        readresult = Btrdfile.gcount();

        //blockread(Corrfile, Corrbuff, thisread, readresult);
        Corrfile.read((char*) &Corrbuff, thisread);
        readresult = Corrfile.gcount();

		readbyteCount = readbyteCount + readresult;
		combinelossybuffers(thisread);
        //BlockWrite(WAVEFile, MergeBuff, thisread, readresult);
        WAVEFile.write((const char*) &MergeBuff, thisread);
        ErrOutput << '\x0d' << "Progress: " << (double(readbyteCount) / (BTRDChunk.WAV.Header.Size + 8) * 100)/*: 0: 2*/ << "%    ";
	}

    Corrfile.close();
    readbyteCount = Btrdfile.tellg();
	readresult = 65536;

	while (readresult == 65536)
	{
        //blockread(Btrdfile, BtrdBuff, 65536, readresult);
        Btrdfile.read((char*) &BtrdBuff, 65536);
        readresult = Btrdfile.gcount();

        //BlockWrite(WAVEFile, BtrdBuff, readresult);
        WAVEFile.write((const char*) &BtrdBuff, readresult);
		readbyteCount = readbyteCount + readresult;
        ErrOutput << '\x0d' << "Progress: " << (double(readbyteCount) / (BTRDChunk.WAV.Header.Size + 8) * 100)/*: 0: 2*/ << "%    ";
	}

    ErrOutput << CRLF;
    Btrdfile.close();
    WAVEFile.close();
}


void readtransfer_two(PByte pB, PByte pEndB)
{
	short* pS;
	int iSample;
	int iChannel;
	int thissample;
	iSample = 0;

	while (pB != pEndB)
	{
		for (iChannel = 0; iChannel < WAVEChunk.FMT.wChannels; ++iChannel)
		{
			pS = reinterpret_cast<short *>(pB);
			pB += 2;
			thissample = (* pS);
			BlockData.Pointer[3][iChannel][iSample] = thissample;
		}

		++ iSample;
	}
}


void readtransfer_three(PByte pB, PByte pEndB)
{
	short* pS;
	int iSample;
	int iChannel;
	int thissample;
	iSample = 0;

	while (pB != pEndB)
	{
		for (iChannel = 0; iChannel < WAVEChunk.FMT.wChannels; ++iChannel)
		{
			pS = reinterpret_cast<short *>(pB + 1);
			thissample = (*pS) * 256;
			thissample |= (*pB);
			pB += 3;
			BlockData.Pointer[3][iChannel][iSample] = thissample;
		}

		++ iSample;
	}
}


void readtransfer_four(PByte pB, PByte pEndB)
{
	int* pI;
	int iSample;
	int iChannel;
	int thissample;
	iSample = 0;

	while (pB != pEndB)
	{
		for (iChannel = 0; iChannel < WAVEChunk.FMT.wChannels; ++iChannel)
		{
			pI = reinterpret_cast<int *>(pB);
			pB += 4;
			thissample = (* pI);
			BlockData.Pointer[3][iChannel][iSample] = thissample;
		}

		++ iSample;
	}
}


bool readNextNextCodecBlock()
{
	bool result = false;
	PByte pB, pStartB;
	unsigned int thisblockreadlength;

	BlockData.Size.Next = 0;

	if (CantReadInWavAnyMore)
	{
		return result;
	}

	if (nrOfBlockInBuffNotYetFetched == 0)  // samples to be fetched aren't in inbuff
	{
		if ((nrOfFullBlocksReadIntoInBuff < nrOfBlocksInReadBuff) || (sampleByteLeftToRead == 0))  // EOF encountered with last call to readNextSampleBlock
		{
			return result; // no more data
		}

		thisblockreadlength = std::min(BUFFER_SIZEread, sampleByteLeftToRead);
		readbyteCount = 0;
		retries = 0;
		readfromWAVEfile(& WAVBuff, thisblockreadlength);
		sampleByteLeftToRead = sampleByteLeftToRead - readbyteCount;
		samplebytesread = samplebytesread + readbyteCount;
		nrOfFullBlocksReadIntoInBuff = readbyteCount / nrOfByteInOneBlockInBuff;
		nrOfByteOfTheLastIncompleteBlock = int(readbyteCount) % -int(nrOfByteInOneBlockInBuff); // why minus here?? // TY
		nrOfBlockInBuffNotYetFetched = 1;
	}

	pB = reinterpret_cast<PByte>(& WAVBuff);
	pB += (nrOfBlockInBuffNotYetFetched - 1) * nrOfByteInOneBlockInBuff;
	pStartB = pB; // points to first sample in Block

	if (nrOfBlockInBuffNotYetFetched <= nrOfFullBlocksReadIntoInBuff)
	{
		pB += nrOfByteInOneBlockInBuff; // pB points to the spot after last sample in Block
		BlockData.Size.Next = nrOfByteInOneBlockInBuff / (((unsigned int) Global.bytes_per_sample * Global.channels));
		nrOfBlockInBuffNotYetFetched = nrOfBlockInBuffNotYetFetched + 1;

		if (nrOfBlockInBuffNotYetFetched > nrOfBlocksInReadBuff)
		{
			nrOfBlockInBuffNotYetFetched = 0;
		}
	}
	else
	{
		if (nrOfByteOfTheLastIncompleteBlock == 0)
		{
			return result;
		}

		pB += nrOfByteOfTheLastIncompleteBlock;  // pB points to the spot after last sample in Block
		BlockData.Size.Next = nrOfByteOfTheLastIncompleteBlock / (((unsigned int) Global.bytes_per_sample * Global.channels));
		nrOfBlockInBuffNotYetFetched = 0; // force a try for a new Blockread with next call which will lead to a return False
	}

	readtransfer(pStartB, pB);
	result = true;
	return result;
}


void WriteTransfer_All(MultiChannelCodecBlock& outputcodecblock, PByte pB)
{
	int wt_i, wt_j;
	int this_sample;

	for (wt_i = 0; wt_i < BlockData.Size.This; ++wt_i)
		for (wt_j = 0; wt_j < WAVEChunk.FMT.wChannels; ++wt_j)
		{
			this_sample = outputcodecblock[wt_j][wt_i];
			(* pB ++) = this_sample & 255;

			if (Global.bytes_per_sample > 1)
			{
				(* pB ++) = (this_sample >>= 8) & 255;

				if (Global.bytes_per_sample > 2)
				{
					(* pB ++) = (this_sample >>= 8) & 255;

					if (Global.bytes_per_sample > 3)
					{
						(* pB ++) = (this_sample >>= 8) & 255;
					}
				}
			}
		}
}


bool writeNextCodecBlock(std::ostream& outputfile, void* OutBuff, MultiChannelCodecBlock& outputcodecblock, unsigned int& FilledOutputBuffer)
{
	bool result;
	unsigned int nrOfByteForOutBuff;
	PByte pB;
	result = false;
	nrOfByteForOutBuff = ((int) Global.bytes_per_sample) * WAVEChunk.FMT.wChannels * BlockData.Size.This;
	writebytecount = 0;

	if (FilledOutputBuffer + nrOfByteForOutBuff > BUFFER_SIZEwrite)
	{
        //BlockWrite(outputfile, OutBuff, FilledOutputBuffer);
        outputfile.write((const char*) &OutBuff, FilledOutputBuffer);

		if (ioresult != 0)
		{
			return result;
		}

		FilledOutputBuffer = 0;
	}

	pB = (PByte) OutBuff; // Check this // TY
	pB += FilledOutputBuffer;
	writetransfer(outputcodecblock, pB);
	FilledOutputBuffer += nrOfByteForOutBuff;
	result = true;
	return result;
}


bool writeNextbtrdcodecblock()
{
	return writeNextCodecBlock(Btrdfile, &BtrdBuff, BlockData.BTRD, nrOfByteFilledInBtrdBuff);
}


bool writeNextcorrcodecblock()
{
	return writeNextCodecBlock(Corrfile, &Corrbuff, BlockData.CORR, nrOfByteFilledInCorrBuff);
}


bool closeWavIO()
{
	bool result = false;

	if (filesOpened > 1)
	{
		if ((filesOpened & 2) == 2 && (nrOfByteFilledInBtrdBuff > 0))
		{
            //BlockWrite(Btrdfile, BtrdBuff, nrOfByteFilledInBtrdBuff);
            Btrdfile.write((const char*) &BtrdBuff, nrOfByteFilledInBtrdBuff);

            if ((Btrdfile.tellg() & 1) == 1) // TY: ?
			{
                //BlockWrite(Btrdfile, Unit_Data_Packer2[0], 1);
                Btrdfile.write((const char*) &Unit_Data_Packer2[0], 1);
			}

			if (ioresult != 0)
			{
				return result;
			}
		}

		if ((filesOpened & 4) == 4 && (nrOfByteFilledInCorrBuff > 0))
		{
            //BlockWrite(Corrfile, Corrbuff, nrOfByteFilledInCorrBuff);
            Corrfile.write((const char*) &Corrbuff, nrOfByteFilledInCorrBuff);

            if ((Corrfile.tellg() & 1) == 1)
			{
                //BlockWrite(Corrfile, Unit_Data_Packer2[0], 1);
                Corrfile.write((const char*) &Unit_Data_Packer2[0], 1);
            }

			if (ioresult != 0)
			{
				return result;
			}
		}

        while (!(WAVEFile.eof()) && (! CantReadInWavAnyMore))
		{
			if ((readfromWAVEfile(& WAVBuff, BUFFER_SIZEread) < ((unsigned int) BUFFER_SIZEread)) || (ioresult != 0))
			{
				CantReadInWavAnyMore = true;
			}

			if (readbyteCount != 0)
			{
				if ((filesOpened & 2) == 2)
				{
                    //BlockWrite(Btrdfile, WAVBuff, readbyteCount);
                    Btrdfile.write((const char*) &WAVBuff, readbyteCount);

					if (ioresult != 0)
					{
						return result;
					}
				}

				if ((filesOpened & 4) == 4)
				{
                    //BlockWrite(Corrfile, WAVBuff, readbyteCount);
                    Corrfile.write((const char*) &WAVBuff, readbyteCount);

					if (ioresult != 0)
					{
						return result;
					}
				}
			}
		}

        WAVEFile.close();

		if (ioresult != 0)
		{
			return result;
		}

		filesOpened = filesOpened ^ 1;

		if ((filesOpened & 2) == 2)
		{
            if ((Btrdfile.tellg() & 1) == 1)
			{
                //BlockWrite(Btrdfile, Unit_Data_Packer2[0], 1);
                Btrdfile.write((const char*) &Unit_Data_Packer2[0], 1);
			}

			if ((parameters.STDINPUT) && (! parameters.STDOUTPUT))
			{
				RewriteDataChunk(Btrdfile);
			}

            Btrdfile.close();

			if (ioresult != 0)
			{
				return result;
			}

			filesOpened = filesOpened ^ 2;
		}

		if ((filesOpened & 4) == 4)
		{
            if ((Corrfile.tellg() & 1) == 1)
			{
                //BlockWrite(Corrfile, Unit_Data_Packer2[0], 1);
                Corrfile.write((const char*) &Unit_Data_Packer2[0], 1);
			}

			if ((parameters.STDINPUT) && (! parameters.STDOUTPUT))
			{
				RewriteDataChunk(Corrfile);
			}

            Corrfile.close();

			if (ioresult != 0)
			{
				return result;
			}

			filesOpened = filesOpened ^ 4;
		}
	}

	result = true;
	return result;
}


bool openWavIO()
{
	bool result = false;
	void (* writetransferprocs[ 5 ])(MultiChannelCodecBlock & outputcodecblock, PByte pB) = { WriteTransfer_All , WriteTransfer_All , WriteTransfer_All, NULL, NULL };
	void (* readtransferprocs[ 5 ])(PByte pB, PByte pEndB) = { readtransfer_two, readtransfer_three, readtransfer_four, NULL, NULL };

	filesOpened = 0;
	Global.Codec_Block_size = 0;
	OpenFile(WAVEFile, parameters.WavInpDir + parameters.wavName, 0, 1);
	ReadChunks(WAVEFile, WAVEChunk);

	if (WAVEChunk.Data.Header.Size == 0)
	{
		wavIOExitProc("Zero sample WAV file!?!", 0x12);
	}

	Global.bits_per_sample = WAVEChunk.FMT.wBitsPerSample;

	if ((Global.bits_per_sample < 9) || (Global.bits_per_sample > 32))
	{
        ErrOutput << "Input file has a bitdepth of " << Global.bits_per_sample << " bits per sample." << CRLF;
		wavIOExitProc("This is not supported.", 0x12);
	}

	Global.bytes_per_sample = (Global.bits_per_sample + 7) >> 3;
	readtransfer = readtransferprocs[Global.bytes_per_sample];
	writetransfer = writetransferprocs[Global.bytes_per_sample];
	Global.channels = WAVEChunk.FMT.wChannels;
	Global.wave_size = WAVEChunk.WAV.Header.Size;
	Global.sample_rate = WAVEChunk.FMT.nSamplesPerSec;
	Global.total_samples = int(1.0 * Global.wave_size / Global.channels / Global.bytes_per_sample);
	Global.sample_rate_recip = 1.0 / Global.sample_rate;

	if (Global.sample_rate < 22050)
	{
        ErrOutput << "Sample Rate too low : " << floattostrf(Global.sample_rate * OneOver[1000], ffFixed, 99, 2) << "kHz (Min=22.05kHz)." << CRLF;
		wavIOExitProc("This is not supported.", 0x12);
	}

	if (Global.sample_rate > 384000)
	{
        ErrOutput << "Sample Rate too high : " << floattostrf(Global.sample_rate * OneOver[1000], ffFixed, 99, 2) << "kHz (Max=384.001kHz)." << CRLF;
		wavIOExitProc("This is not supported.", 0x12);
	}

	Global.Codec_Block_duration = OneOver[100];
	Global.Codec_Block_bits = std::max(8, nround(nlog2(Global.Codec_Block_duration * Global.sample_rate)));
	Global.Codec_Block_size = PowersOf.TwoInt[Global.Codec_Block_bits];
	Global.Codec_Block_bit_shift = Global.Codec_Block_bits - 9;
	Global.Codec_Block_duration = double(Global.Codec_Block_size) / Global.sample_rate;

	if (Global.Codec_Block_bits > MAX_FFT_BIT_LENGTH - 1)
	{
        ErrOutput << "Sample Rate too high : " << floattostrf(Global.sample_rate * OneOver[1000], ffFixed, 99, 1) << "kHz (Max=409.6kHz)." << CRLF;
	}

	sampleByteLeftToRead = WAVEChunk.Data.Header.Size;
	nrOfByteInOneBlockInBuff = Global.bytes_per_sample * Global.channels * Global.Codec_Block_size;

	if (parameters.STDINPUT)
	{
		nrOfBlocksInReadBuff = 1;
	}
	else
	{
		nrOfBlocksInReadBuff = (BUFFER_SIZE / nrOfByteInOneBlockInBuff);
	}

	if (parameters.STDOUTPUT)
	{
		nrOfBlocksInWriteBuff = 1;
	}
	else
	{
		nrOfBlocksInWriteBuff = (BUFFER_SIZE / nrOfByteInOneBlockInBuff);
	}

	BUFFER_SIZEread = nrOfByteInOneBlockInBuff * nrOfBlocksInReadBuff;
	BUFFER_SIZEwrite = nrOfByteInOneBlockInBuff * nrOfBlocksInWriteBuff;
	CantReadInWavAnyMore = false;

	if (parameters.checking)
	{
		if ((WAVEChunk.FACT.Header.Size > 0) && (std::string(&WAVEChunk.FACT.Data[0], 8) == "lossyWAV"))
		{
			ExitCode = 0x10;

			if ((parameters.verbosity) && (! parameters.silent))
			{
                ErrOutput << "lossyWAV FACT Chunk found. File already processed." << CRLF;
                ErrOutput << WAVEChunk.FACT.Data;
			}
		}
		else
		{
			ExitCode = 0x00;

			if ((parameters.verbosity) && (! parameters.silent))
			{
                ErrOutput << "lossyWAV FACT Chunk not found. File not marked as processed." << CRLF;
			}
		}

		closeWavIO();
		throw (ExitCode);
	}
	else
	{
		if (std::string(&WAVEChunk.FACT.Data[0], 8) == "lossyWAV")
		{
            ErrOutput << WAVEChunk.FACT.Data;
			lossyWAVerror("lossyWAV FACT Chunk found. File already processed.", 0x10);
		}
	}

	strings.datestamp = std::ctime(&timer.StartTime);
	std::memcpy(WAVEChunk.FACT.Header.id, "fact", 4);
    DateTimeString = version_string + strings.version_short + " @ " + strings.datestamp + ", " + strings.parameter + "\n\x00";

    if ((DateTimeString.length() & 1) == 1)
	{
		DateTimeString = DateTimeString + "\x00";
	}

	std::memcpy(&WAVEChunk.FACT.Data, &DateTimeString[0], DateTimeString.length()); // NB! index 0 is first char
	WAVEChunk.FACT.Header.Size = DateTimeString.length();

	if (WAVEChunk.WAV.Header.Size < MAX_WAVE_SIZE)
	{
		WAVEChunk.WAV.Header.Size += WAVEChunk.FACT.Header.Size + 8;
	}

	firstsamplesread = std::min(sampleByteLeftToRead, BUFFER_SIZEread);

	if ((readfromWAVEfile(& WAVBuff, firstsamplesread) < firstsamplesread) || (CantReadInWavAnyMore))
	{
		CantReadInWavAnyMore = true;
		return result;
	}

	samplebytesread = readbyteCount;
	sampleByteLeftToRead = sampleByteLeftToRead - readbyteCount;
	nrOfFullBlocksReadIntoInBuff = readbyteCount / nrOfByteInOneBlockInBuff;
	nrOfByteOfTheLastIncompleteBlock = readbyteCount - nrOfFullBlocksReadIntoInBuff * nrOfByteInOneBlockInBuff;
	nrOfBlockInBuffNotYetFetched = 1;
	nrOfByteFilledInBtrdBuff = 0;
	nrOfByteFilledInCorrBuff = 0;
	OpenFile(Btrdfile, parameters.WavOutDir + parameters.lossyName, 2, 2);

	if (! WriteChunks(Btrdfile))
	{
		lossyWAVerror("Writing to output file.", 0x12);
	}

	if ((! parameters.STDOUTPUT) && (parameters.correction))
	{
		OpenFile(Corrfile, parameters.WavOutDir + parameters.lwcdfName, 2, 4);

		if (! WriteChunks(Corrfile))
		{
			lossyWAVerror("Writing to output file.", 0x12);
		}
	}

	result = true;
	return result;
}


void nWAV_Init()
{
	int co_i;
	chunkspace = new unsigned char[ chunkspace_array_size ];

	for (co_i = 0; co_i < MAX_CHANNELS; ++co_i)
	{
		BlockData.CHANNEL_OFFSET[co_i] = co_i * CHANNEL_BYTE_SIZE;
	}
}


static struct Init
{
	Init() {
		// initialization: todo
		Unit_Data_Packer2[0] = 0;
		nWAV_Init();
	}
	~Init() {
		// finalization: todo Write ( Erroutput, UnitName, CRLF );
		delete[] chunkspace;
	}
} init;
